;;; term-nasty.el --- Damned Things from terminfo.el
;;; This file is in the public domain, and was written by Stallman and Mlynarik

;;; Commentary:

;; Some people used to be bothered by the following comments that were
;; found in terminal.el.  We decided they were distracting, and that it
;; was better not to have them there.  On the other hand, we didn't want
;; to appear to be giving in to the pressure to censor obscenity that
;; currently threatens freedom of speech and of the press in the US.
;; So we decided to put the comments here.

;;; Code:

These comments were removed from te-losing-unix.
  ;(what lossage)
  ;(message "fucking-unix: %d" char)

This was before te-process-output.
;; fucking unix has -such- braindamaged lack of tty control...

And about the need to handle output characters such as C-m, C-g, C-h
and C-i even though the termcap doesn't say they may be used:
;fuck me harder
;again and again!
;wa12id!!
;(spiked)

;;; term-nasty.el ends here
